@@ -8,6 +8,8 @@ use warnings;
 
 sub D_PATTERN () { 0 }
 sub D_RULES () { 0 }
+sub D_EQUIV () { 0 }
+sub D_TRACE () { 0 }
 
 # Preamble
 
@@ -16,12 +18,12 @@ require Exporter;
 our @ISA = qw(Exporter);
 
 our @EXPORT = qw( declare rule unknown stringify
-                  capture is_num pattern
-		  $hold
+                  capture is_num pattern equiv
+		  is_atom
 		);
 for (0..9) { push @EXPORT, "_$_" } # all the shortened capture sequences
 
-our $VERSION = '0.00102'; # increment _after_ each distribution
+our $VERSION = '0.00103'; # increment _after_ each distribution
 
 use Carp;
 use YAML;
@@ -29,6 +31,8 @@ use Scalar::Util qw( blessed );
 
 # Code below
 
+our @captures;
+
 { # quick'n'dirty wrapper
   package Grapefruit::Rule;
   sub new {
@@ -56,7 +60,7 @@ use Scalar::Util qw( blessed );
     my ($func, @args) = @_;
     croak "Function name needs to be a simple scalar" if ref $func;
     my $self = [$func, @args];
-    return bless $self, $class
+    return bless $self, $class;
   }
   # this is a method, but it calls the multimethod...
   sub stringify {
@@ -64,20 +68,65 @@ use Scalar::Util qw( blessed );
     #print "$self->[0] ".join(" ", @$self[1..$#$self])."\n";
     "$self->[0]( ".join(", ", map Grapefruit::stringify($_), @$self[1..$#$self])." )";
   }
+  sub _do_match {
+    my ($self, $parsetree) = @_;
+    if (UNIVERSAL::isa($parsetree, 'Grapefruit::Compound')) {
+      return Grapefruit::_do_match_apply($self, $parsetree);
+    } else {
+      return 0;
+    }
+  }
+  sub equiv {
+    my ($self, $right) = @_;
+    if (UNIVERSAL::isa($right, 'Grapefruit::Compound')) {
+      return 0 unless $self->[0] eq $right->[0];
+      return 0 unless @$self == @$right;
+      for (1..$#$self) {
+	return 0 unless Grapefruit::equiv($self->[$_], $right->[$_]);
+      }
+      return 1;
+    } else {
+      return 0;
+    }
+  }
 }
 
 { # quick'n'dirty wrapper
   package Grapefruit::Pattern;
+  BEGIN { *D_PATTERN = *Grapefruit::D_PATTERN };
   sub new {
     my $proto = shift;
     my $class = ref $proto || $proto;
-    my ($patterntree) = @_;
-    my $self = sub {};
+    my ($patterntree, $predicate) = @_;
+
+    my $self = sub {
+      my ($parsetree) = @_;
+      local @captures = ();
+      D_PATTERN and print "* Patterntree: $patterntree\n";
+      D_PATTERN and print "* Parsetree: $parsetree\n";
+
+      # recursivly process the two trees in parallel
+      my $success = Grapefruit::_do_match($patterntree, $parsetree);
+
+      # check the predicate
+      if (defined $predicate) {
+        $success &&= $predicate->(@captures);
+      }
+    
+      D_PATTERN and print "* Captures: @captures\n";
+      D_PATTERN and print "* Success: $success\n";
+    
+      if ($success) {
+        return wantarray ? @captures : 1;
+      } else {
+        return wantarray ? () : 0;
+      }
+    };
+
     return bless $self, $class;
   }
 }
 
-our @captures;
 sub capture ($;$) {
   my ($num, $matcher) = @_;
   my $closure;
@@ -130,7 +179,9 @@ sub _do_match_apply {
 sub _do_match {
   my ($patterntree, $parsetree) = @_;
   D_PATTERN and print "** _do_match $patterntree $parsetree\n";
-  if (ref $patterntree eq "ARRAY" and
+  if (blessed $patterntree and $patterntree->can('_do_match')) {
+    $patterntree->_do_match($parsetree);
+  } elsif (ref $patterntree eq "ARRAY" and
       ref $parsetree eq "ARRAY") {
     return _do_match_apply($patterntree, $parsetree);
   } elsif (ref $patterntree eq "pattern::capture") {
@@ -152,29 +203,10 @@ sub _do_match {
 }
 
 # take a parsetree and generate a matcher
-sub pattern ($) {
-  my ($patterntree) = @_;
+sub pattern ($;$) {
+  my ($patterntree, $pred) = @_;
 
-  my $closure = sub {
-    my ($parsetree) = @_;
-    local @captures = ();
-    D_PATTERN and print "* Patterntree: $patterntree\n";
-    D_PATTERN and print "* Parsetree: $parsetree\n";
-
-    # recursivly process the two trees in parallel
-    my $success = _do_match($patterntree, $parsetree);
-    
-    D_PATTERN and print "* Captures: @captures\n";
-    D_PATTERN and print "* Success: $success\n";
-    
-    if ($success) {
-      return wantarray ? @captures : 1;
-    } else {
-      return wantarray ? () : 0;
-    }
-  };
-
-  return bless $closure, 'pattern';
+  return Grapefruit::Pattern->new($patterntree, $pred);
 }
 
 our %rules;
@@ -193,19 +225,27 @@ sub run_rules ($;$) {
 
   my @levels = sort keys %rules;
   @levels = grep { $_ <= $prec } @levels if defined $prec;
-  D_RULES and print "levels: @levels\n";
+  D_RULES and print "levels: @levels (applying to $expr)\n";
 
   LEVEL: for my $level (@levels) {
     D_RULES and print "level: $level\n";
     RULE: for my $rule (@{$rules{$level}}) {
-      D_RULES and print "  rule: $rule\n";
+      D_RULES and print "  rule: $rule (applied to $expr)\n";
       if (my @captures = $rule->match($expr)) {
-	local $hold = 1;
+	#local $hold = 1; # disabled
+
 	D_RULES and print "    match!\n";
 	D_RULES and print "Pre-transform:\n", Dump($expr);
 	D_RULES and print "Captures: @captures\n";
+
+	D_TRACE and print ": ", stringify($expr), "\n";
+
 	$expr = $rule->transform(@captures);
+
+	D_TRACE and print "  --> ", stringify($expr), "\n";
+
 	D_RULES and print "Post-transform:\n", Dump($expr);
+
 	# now what do we skip to the next precedence level or restart this one?
 	# last LEVEL; # terminate early
 	# next LEVEL; # skip to the next one
@@ -226,7 +266,7 @@ sub declare ($) {
   my $code = <<"END";
     package $pkg;
     sub $name {
-      Grapefruit::run_rules Grapefruit::Compound->new(q{$name}, \@_);
+      Grapefruit::run_rules(Grapefruit::Compound->new(q{$name}, \@_));
     }
 END
   eval $code;
@@ -252,6 +292,26 @@ sub is_num ($) {
   return $_[0] =~ /^\d+$/; # only handles integers, see perlfaq for more REs
 }
 
+sub is_atom ($) {
+  return 1 if UNIVERSAL::isa($_[0], 'symbolic::unknown');
+  return ref $_[0] ? 0 : 1;
+}
+
+# bah! It'll work!
+sub equiv {
+  my ($left, $right) = @_;
+  D_EQUIV and printf "equiv: [%s] [%s]\n", stringify($left), stringify($right);
+  if (blessed $left and $left->can('equiv')) {
+    return $left->equiv($right);
+  } elsif (blessed $right and $right->can('equiv')) {
+    return $right->equiv($left);
+  } elsif (is_num $left and is_num $right) {
+    return $left == $right;
+  } else {
+    return $left eq $right;
+  }
+}
+
 1;
 __END__
 
